import json


def load_data(data_path):
    """
    Load data from a JSON file
    the data should be loaded here with the json path and return it to process the data
    """
    with open(data_path) as file:
        data = json.load(file)
    return data



import numpy as np
from tensorflow import keras
from nltk.stem.lancaster import LancasterStemmer
import nltk
import os
import pickle
from neurolink.include import __load__
from neurolink.include.__process__ import *


# Build and compile the model
def build_model(input_shape, output_shape):
    model = keras.Sequential([
        keras.layers.Dense(8, input_shape=input_shape, activation='relu'),
        keras.layers.Dense(output_shape, activation='softmax')
    ])
    model.compile(optimizer='adam', loss='categorical_crossentropy', metrics=['accuracy'])
    return model


# Train the model
def train_model(model, training, output, epochs, verbose=1):
    model.fit(training, output, epochs=epochs, verbose=verbose)


# Predict the intent
def predict_intent(sentence, model, words, labels):
    sentence_words = nltk.word_tokenize(sentence)
    sentence_words = [LancasterStemmer().stem(word.lower()) for word in sentence_words]
    input_bag = [0 for _ in range(len(words))]
    for word in sentence_words:
        for i, w in enumerate(words):
            if w == word:
                input_bag[i] = 1
    input_data = np.array(input_bag).reshape(1, -1)
    prediction = model.predict(input_data)[0]
    intent_index = np.argmax(prediction)
    intent = labels[intent_index]
    return intent


# Use the model to predict intents
def chat(message, confidence_threshold, intents_path, model_path, data_path, should_train_model):
    if should_train_model:
        # Load and preprocess the data
        data = __load__.load_data(intents_path)
        words, labels, training, output = preprocess_data(data)

        # Build and compile the model
        input_shape = (len(training[0]),)
        output_shape = len(output[0])
        model = build_model(input_shape, output_shape)

        # Train the model
        epochs = 1000
        train_model(model, training, output, epochs, verbose=0)

        # Save the model
        model.save(model_path)

        # Save words and labels using pickle
        with open(data_path, 'wb') as file:
            pickle.dump((words, labels), file)

    else:
        model = keras.models.load_model(model_path)
        with open(data_path, 'rb') as file:
            words, labels = pickle.load(file)

    response = predict_intent(message, model, words, labels)
    if response:
        return response
    else:
        return "Sorry, I don't understand. Please try again."


class botModel_keras:
    """
    these are the test case for the chat function
intents_path = os.path.abspath("intents.json")
model_path = os.path.abspath("model.tflearn")
data_path = os.path.abspath("data.pickle")

response = chat("Hi", 0.5, intents_path, model_path, data_path, train_model=False)
print(response)
"""

    def __init__(self, intents_path, model_path, data_path, train_model):
        self.intents_path = intents_path
        self.model_path = model_path
        self.data_path = data_path
        self.train_model = train_model

    def chat(self, message, accuracy):
        response = chat(message, accuracy, self.intents_path, self.model_path, self.data_path, self.train_model)
        return response



import os
import random
import json
import nltk
import numpy as np
import torch
import torch.nn as nn
from nltk import LancasterStemmer
from torch.utils.data import Dataset, DataLoader
from nltk.stem.porter import PorterStemmer

from neurolink.include import __load__

stemmer = PorterStemmer()

model_path = os.path.abspath("../../intents.json")

intents = __load__.load_data(model_path)


def bag_of_words(tokenized_sentence, words):
    # Stem each word
    sentence_words = [stem(word) for word in tokenized_sentence]
    # Initialize bag with 0 for each word
    bag = np.zeros(len(words), dtype=np.float32)
    for idx, w in enumerate(words):
        if w in sentence_words:
            bag[idx] = 1
    return bag


def tokenize(sentence):
    return nltk.word_tokenize(sentence)


def stem(word):
    return stemmer.stem(word.lower())


class ChatDataset(Dataset):
    def __init__(self, X, y):
        self.n_samples = len(X)
        self.x_data = X
        self.y_data = y

    def __getitem__(self, index):
        return self.x_data[index], self.y_data[index]

    def __len__(self):
        return self.n_samples


class NeuralNet(nn.Module):
    def __init__(self, input_size, hidden_size, num_classes):
        super(NeuralNet, self).__init__()
        self.l1 = nn.Linear(input_size, hidden_size)
        self.l2 = nn.Linear(hidden_size, hidden_size)
        self.l3 = nn.Linear(hidden_size, num_classes)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = x.to(torch.float32)  # Convert x to float32 data type
        out = self.l1(x)
        out = self.relu(out)
        out = self.l2(out)
        out = self.relu(out)
        out = self.l3(out)
        return out



def __preprocess_data__(intents):
    all_words = []
    tags = []
    xy = []

    for intent in intents['intents']:
        tag = intent['tag']
        tags.append(tag)
        for pattern in intent['patterns']:
            w = tokenize(pattern)
            all_words.extend(w)
            xy.append((w, tag))

    ignore_words = ['?', '.', '!']
    all_words = [stem(w) for w in all_words if w not in ignore_words]
    all_words = sorted(set(all_words))
    tags = sorted(set(tags))

    X_train = []
    y_train = []
    for (pattern_sentence, tag) in xy:
        bag = bag_of_words(pattern_sentence, all_words)
        X_train.append(bag)
        label = tags.index(tag)
        y_train.append(label)

    X_train = np.array(X_train)
    y_train = np.array(y_train)

    return X_train, y_train, all_words, tags


X_train, y_train, all_words, tags = __preprocess_data__(intents)

num_epochs = 1000
batch_size = 8
learning_rate = 0.001
input_size = len(X_train[0])
hidden_size = 8
output_size = len(tags)
print(input_size, output_size)


def train():
    dataset = ChatDataset(X_train, y_train)
    train_loader = DataLoader(dataset=dataset, batch_size=batch_size, shuffle=True, num_workers=0)
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    model = NeuralNet(input_size, hidden_size, output_size).to(device)
    criterion = nn.CrossEntropyLoss()
    optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)

    for epoch in range(num_epochs):
        for (words, labels) in train_loader:
            words = words.to(device)
            labels = labels.to(dtype=torch.long).to(device)
            optimizer.zero_grad()
            outputs = model(words)
            loss = criterion(outputs, labels)
            loss.backward()
            optimizer.step()

        if (epoch + 1) % 100 == 0:
            print(f'Epoch [{epoch + 1}/{num_epochs}], Loss: {loss.item():.4f}')

    print(f'Final loss: {loss.item():.4f}')

    data = {
        "model_state": model.state_dict(),
        "input_size": input_size,
        "hidden_size": hidden_size,
        "output_size": output_size,
        "all_words": all_words,
        "tags": tags
    }
    data_path = os.path.abspath("data.pth")
    FILE = data_path
    torch.save(data, FILE)
    print(f'Training complete. File saved to {FILE}')


def chat(message, threshold, intents, data_path):
    # Load other data
    data = torch.load(data_path)
    input_size = data["input_size"]
    hidden_size = data["hidden_size"]
    output_size = data["output_size"]
    all_words = data['all_words']
    tags = data['tags']
    model_state = data["model_state"]
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    model = NeuralNet(input_size, hidden_size, output_size).to(device)
    model.load_state_dict(model_state)
    model.eval()
    sentence = message.lower()
    sentence = tokenize(sentence)
    X = bag_of_words(sentence, all_words)
    X = X.reshape(1, X.shape[0])
    X = torch.from_numpy(X).to(device)
    output = model(X)
    _, predicted = torch.max(output, dim=1)
    tag = tags[predicted.item()]
    probs = torch.softmax(output, dim=1)
    prob = probs[0][predicted.item()]
    if prob.item() > threshold:
        for intent in intents:
            if tag == intent["tag"]:
                return random.choice(intent['responses'])
    else:
        return "I do not understand... Could you please rephrase?"


train_model = True

if train_model:
    train()

intents_path = os.path.abspath("../../intents.json")
data_path = os.path.abspath("data.pth")
threshold = 0.75

# Load intents file
with open(intents_path, 'r') as file:
    intents = json.load(file)["intents"]

while True:
    message = input("User: ")
    if message.lower() == "quit":
        break
    response = chat(message, threshold, intents, data_path)
    print("Assistant:", response)



import nltk
import numpy as np
from nltk import LancasterStemmer



def preprocess_data(data):
    words = []
    labels = []
    docs_x = []
    docs_y = []
    for intent in data["intents"]:
        for pattern in intent["patterns"]:
            wrds = nltk.word_tokenize(pattern)
            words.extend(wrds)
            docs_x.append(wrds)
            docs_y.append(intent["tag"])
        if intent["tag"] not in labels:
            labels.append(intent["tag"])
    words = [LancasterStemmer().stem(w.lower()) for w in words if w != "?"]
    words = sorted(list(set(words)))
    labels = sorted(labels)
    training = []
    output = []
    out_empty = [0 for _ in range(len(labels))]
    for x, doc in enumerate(docs_x):
        bag = []
        wrds = [LancasterStemmer().stem(w) for w in doc]
        for w in words:
            if w in wrds:
                bag.append(1)
            else:
                bag.append(0)
        output_row = out_empty[:]
        output_row[labels.index(docs_y[x])] = 1
        training.append(bag)
        output.append(output_row)
    training = np.array(training)
    output = np.array(output)
    return words, labels, training, output




import os

from neurolink.Scripts.__tflearn__ import chat


class initialize:
    """
    you can pass the data in this format => path of the intents.json, model.tflearn, data.pickle and train_model as
    True or False default is True
    intents_path = os.path.abspath("intents.json")
    model_path = os.path.abspath("model.tflearn")
    data_path = os.path.abspath("data.pickle")
    initialize_obj = initialize(intents_path, model_path, data_path, train_model=True)
    response = initialize_obj.chat("Hi", 0.5)
    print(response)

    """
    def __init__(self, intents_path, model_path, data_path, train_model=True):
        self.intents_path = intents_path
        self.model_path = model_path
        self.data_path = data_path
        self.train_model = train_model

    def chat(self, message, accuracy):
        response = chat(message, accuracy, self.intents_path, self.model_path, self.data_path, self.train_model)
        return response





import os
import pickle
import random
import warnings

import nltk
import numpy as np
import tensorflow as tf
import tflearn
from nltk import LancasterStemmer

from neurolink.include.__load__ import load_data
from neurolink.include.__process__ import preprocess_data

warnings.filterwarnings("ignore")



def load_model(model_path, training, output):
    tf.compat.v1.reset_default_graph()
    net = tflearn.input_data(shape=[None, len(training[0])])
    net = tflearn.fully_connected(net, 8)
    net = tflearn.fully_connected(net, 8)
    net = tflearn.fully_connected(net, len(output[0]), activation="softmax")
    net = tflearn.regression(net)
    model = tflearn.DNN(net)
    model.load(model_path)
    return model


def classify_input(sentence, words, labels, model):
    sentence_words = nltk.word_tokenize(sentence)
    sentence_words = [LancasterStemmer().stem(word.lower()) for word in sentence_words]
    input_bag = [0 for _ in range(len(words))]
    for word in sentence_words:
        for i, w in enumerate(words):
            if w == word:
                input_bag[i] = 1
    results = model.predict([input_bag])[0]
    results_index = np.argmax(results)
    tag = labels[results_index]
    return tag, results


def get_response(intents, intent, user_input):
    for i in intents:
        if i["tag"] == intent:
            result = random.choice(i["responses"])
            break
        else:
            result = "Sorry, I don't understand. Please try again."
    return result


def chat(message, accuracy, intents_path, model_path, data_path, train_model):
    # Load the intents data
    intents_data = load_data(intents_path)

    # Preprocess the data
    words, labels, training, output = preprocess_data(intents_data)

    if train_model:
        # Train the model or load the saved model
        tf.compat.v1.reset_default_graph()
        net = tflearn.input_data(shape=[None, len(training[0])])
        net = tflearn.fully_connected(net, 8)
        net = tflearn.fully_connected(net, 8)
        net = tflearn.fully_connected(net, len(output[0]), activation="softmax")
        net = tflearn.regression(net)
        model = tflearn.DNN(net)
        model.fit(training, output, n_epoch=1000, batch_size=8, show_metric=True)
        model.save(model_path)
        with open(data_path, "wb") as f:
            pickle.dump((words, labels, training, output), f)
    else:
        # Load the saved model
        model = load_model(model_path, training, output)

    # Classify the user input
    tag, results = classify_input(message.lower(), words, labels, model)

    # Generate a response based on the predicted intent
    if results[results.argmax()] > accuracy:
        response = get_response(intents_data["intents"], tag, message.lower())
    else:
        # Generate predicted sentence
        predicted_sentence = predict(message, model, words)
        response = "I'm sorry, I'm not sure I understand what you are asking. Could you please provide me " \
                   "with more details or clarify your question? Predicted Sentence: " + predicted_sentence

    return response


def predict(sentence, model, words):
    input_bag = [0 for _ in range(len(words))]
    sentence_words = nltk.word_tokenize(sentence)
    sentence_words = [LancasterStemmer().stem(word.lower()) for word in sentence_words]
    for word in sentence_words:
        for i, w in enumerate(words):
            if w == word:
                input_bag[i] = 1
    input_data = np.array(input_bag).reshape(1, -1)
    prediction = model.predict(input_data)[0]
    predicted_sentence = " ".join([words[i] for i, val in enumerate(prediction) if val > 0.5])
    return predicted_sentence


class initialize:
    """
    these are the test case for the chat function
intents_path = os.path.abspath("intents.json")
model_path = os.path.abspath("model.tflearn")
data_path = os.path.abspath("data.pickle")

response = chat("Hi", 0.5, intents_path, model_path, data_path, train_model=False)
print(response)
"""

    def __init__(self, intents_path, model_path, data_path, train_model):
        self.intents_path = intents_path
        self.model_path = model_path
        self.data_path = data_path
        self.train_model = train_model

    def chat(self, message, accuracy):
        response = chat(message, accuracy, self.intents_path, self.model_path, self.data_path, self.train_model)
        return response



from setuptools import setup, find_packages

VERSION = '0.0.0'
DESCRIPTION = 'NeuroLink is a Python package in active development that aims to provide various AI capabilities, ' \
              'including a chatbot, image processing, audio processing, Django & Flask support, and more.'

with open("README.md", "r", encoding="utf-8") as fh:
    LONG_DESCRIPTION = fh.read()

# Setting up
setup(
    name="neurolink",
    version=VERSION,
    author="Muhammad Fiaz",
    author_email="",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url='https://github.com/muhammad-fiaz/neurolink',
    packages=find_packages(),
    keywords=['neural networks', 'neurolink', 'machine learning', 'artificial intelligence', 'ai', 'ml', 'chatbot',
              'image processing', 'audio processing', 'django', 'flask'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    python_requires='>=3.8',
    install_requires=[
        'pip==23.1.2',
        'attrs',
        'argparse',
        'wheel',
        'packages',
        'google',
        'cryptography',
        'Pillow',
        'docutils',
        'Jinja2',
        'keyring',
        'setuptools',
        'packaging',
        'pyparsing',
        'zipp',
        'nltk',
        'numpy',
        'tensorflow',
        'tflearn',
        'torch'
    ],
    license='Apache Software License (Apache 2.0)',  # License metadata
    # Tags metadata
    project_urls={
        'Source Code': 'https://github.com/muhammad-fiaz/neurolink',
        'Bug Tracker': 'https://github.com/muhammad-fiaz/neurolink/issues',
        'Documentation': 'https://github.com/muhammad-fiaz',

    },
)

print("Happy Coding!")
